#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------

from copy import deepcopy
import os
import warnings
import numpy as np
import pandas as pds

import pysat
import pysat.utils._core as core_utils
from pysat.utils import testing


class Meta(object):
    """ Stores metadata for Instrument instance, similar to CF-1.6 netCDFdata
    standard.

    Parameters
    ----------
    metadata : pandas.DataFrame
        DataFrame should be indexed by variable name that contains at minimum
        the standard_name (name), units, and long_name for the data stored in
        the associated pysat Instrument object.
    labels : dict
        Dict where keys are the label attribute names and the values are tuples
        that have the label values and value types in that order.
        (default={'units': ('units', str), 'name': ('long_name', str),
        'notes': ('notes', str), 'desc': ('desc', str),
        'min_val': ('value_min', float), 'max_val': ('value_max', float),
        'fill_val': ('fill', float)})
    export_nan : list or NoneType
        List of labels that should be exported even if their value is nan or
        None for an empty list. When used, metadata with a value of nan will
        be excluded from export. Will always allow nan export for labels of
        the float type (default=None)

    Attributes
    ----------
    data : pandas.DataFrame
        index is variable standard name, 'units', 'long_name', and other
        defaults are also stored along with additional user provided labels.
    labels : MetaLabels
        Labels for MetaData attributes

    Note
    ----
    Meta object preserves the case of variables and attributes as it first
    receives the data. Subsequent calls to set new metadata with the same
    variable or attribute will use case of first call. Accessing or setting
    data thereafter is case insensitive. In practice, use is case insensitive
    but the original case is preserved. Case preseveration is built in to
    support writing files with a desired case to meet standards.

    Metadata for higher order data objects, those that have
    multiple products under a single variable name in a pysat.Instrument
    object, are stored by providing a Meta object under the single name.

    Supports any custom metadata values in addition to the expected metadata
    attributes (units, name, notes, desc, value_min, value_max, and fill).
    These base attributes may be used to programatically access and set types
    of metadata regardless of the string values used for the attribute. String
    values for attributes may need to be changed depending upon the standards
    of code or files interacting with pysat.

    Meta objects returned as part of pysat loading routines are automatically
    updated to use the same values of units, etc. as found in the
    pysat.Instrument object.

    Examples
    --------
    ::

        # instantiate Meta object, default values for attribute labels are used
        meta = pysat.Meta()
        # set a couple base units
        # note that other base parameters not set below will
        # be assigned a default value
        meta['name'] = {'long_name': string, 'units': string}
        # update 'units' to new value
        meta['name'] = {'units': string}
        # update 'long_name' to new value
        meta['name'] = {'long_name': string}
        # attach new info with partial information, 'long_name' set to 'name2'
        meta['name2'] = {'units': string}
        # units are set to '' by default
        meta['name3'] = {'long_name': string}

        # assigning custom meta parameters
        meta['name4'] = {'units': string, 'long_name': string
                         'custom1': string, 'custom2': value}
        meta['name5'] = {'custom1': string, 'custom3': value}

        # assign multiple variables at once
        meta[['name1', 'name2']] = {'long_name': [string1, string2],
                                    'units': [string1, string2],
                                    'custom10': [string1, string2]}

        # assiging metadata for n-Dimensional variables
        meta2 = pysat.Meta()
        meta2['name41'] = {'long_name': string, 'units': string}
        meta2['name42'] = {'long_name': string, 'units': string}
        meta['name4'] = {'meta': meta2}

        # or
        meta['name4'] = meta2
        meta['name4'].children['name41']

        # mixture of 1D and higher dimensional data
        meta = pysat.Meta()
        meta['dm'] = {'units': 'hey', 'long_name': 'boo'}
        meta['rpa'] = {'units': 'crazy', 'long_name': 'boo_whoo'}
        meta2 = pysat.Meta()
        meta2[['higher', 'lower']] = {'meta': [meta, None],
                                      'units': [None, 'boo'],
                                      'long_name': [None, 'boohoo']}

        # assign from another Meta object
        meta[key1] = meta2[key2]

        # access fill info for a variable, presuming default label
        meta[key1, 'fill']

        # access same info, even if 'fill' not used to label fill values
        meta[key1, meta.fill_label]


        # change a label used by Meta object
        meta.labels.fill_val = '_FillValue'

        # Note that the fill label is intended for use when interacting
        # with external files. Thus, any fill values (NaN) within the Meta
        # object are not updated when changing the metadata string label,
        # or when updating the value representing fill data. A future update
        # (Issue #707) will expand functionality to include these custom
        # fill values when producing files.

    """

    # -----------------------------------------------------------------------
    # Define the magic methods

    def __init__(self, metadata=None,
                 labels={'units': ('units', str), 'name': ('long_name', str),
                         'notes': ('notes', str), 'desc': ('desc', str),
                         'min_val': ('value_min', float),
                         'max_val': ('value_max', float),
                         'fill_val': ('fill', float)}, export_nan=None):

        # Set mutability of Meta attributes.  This flag must be set before
        # anything else, or `__setattr__` breaks.
        self.mutable = True

        # Set the NaN export list
        self._export_nan = [] if export_nan is None else export_nan
        for lvals in labels.values():
            if lvals[0] not in self._export_nan and lvals[1] == float:
                self._export_nan.append(lvals[0])

        # Set the labels
        self.labels = MetaLabels(metadata=self, **labels)

        # init higher order (nD) data structure container, a dict
        self._ho_data = {}

        # Use any user provided data to instantiate object with data
        # Attributes unit and name labels are called within
        if metadata is not None:
            if isinstance(metadata, pds.DataFrame):
                self._data = metadata

                # Make sure defaults are taken care of for required metadata
                self.accept_default_labels(self)
            else:
                raise ValueError(''.join(('Input must be a pandas DataFrame',
                                          'type. See other constructors for',
                                          ' alternate inputs.')))
        else:
            columns = [getattr(self.labels, mlab)
                       for mlab in self.labels.label_type.keys()]
            self._data = pds.DataFrame(None, columns=columns)

        # Establish attributes intrinsic to object, before user can
        # add any
        self._base_attr = dir(self)

    def __repr__(self):
        """String describing MetaData instantiation parameters

        Returns
        -------
        out_str : str
            Simply formatted output string

        """
        nvar = len([kk for kk in self.keys()])
        out_str = ''.join(['pysat.Meta(metadata=', self._data.__repr__(),
                           ', labels=', self.labels.__repr__(),
                           'export_nan=', self._export_nan.__repr__(),
                           ') -> {:d} Variables'.format(nvar)])
        return out_str

    def __str__(self, long_str=True):
        """String describing Meta instance, variables, and attributes

        Parameters
        ----------
        long_str : bool
            Return short version if False and long version if True
            (default=True)

        Returns
        -------
        out_str : str
            Nicely formatted output string

        """
        # Get the desired variables as lists
        labs = [var for var in self.attrs()]
        vdim = [var for var in self.keys() if var not in self.keys_nD()]
        nchild = {var: len([kk for kk in self[var]['children'].keys()])
                  for var in self.keys_nD()}
        ndim = ["{:} -> {:d} children".format(var, nchild[var])
                for var in self.keys_nD()]

        # Get the lengths of each list
        nlabels = len(labs)
        nvdim = len(vdim)
        nndim = len(ndim)

        # Print the short output
        out_str = "pysat Meta object\n"
        out_str += "-----------------\n"
        out_str += "Tracking {:d} metadata values\n".format(nlabels)
        out_str += "Metadata for {:d} standard variables\n".format(nvdim)
        out_str += "Metadata for {:d} ND variables\n".format(nndim)

        # Print the longer output
        if long_str:
            # Print all the metadata labels
            out_str += "\n{:s}".format(self.labels.__str__())

            # Print a subset of the metadata variables, divided by order
            ncol = 3
            max_num = 6  # Should be divible by 2 and ncol
            if nvdim > 0:
                out_str += "\nStandard Metadata variables:\n"
                out_str += core_utils.fmt_output_in_cols(vdim, ncols=ncol,
                                                         max_num=max_num)
            if nndim > 0:
                out_str += "\nND Metadata variables:\n"
                out_str += core_utils.fmt_output_in_cols(ndim, ncols=ncol,
                                                         max_num=max_num)

        return out_str

    def __setattr__(self, name, value):
        """Conditionally sets attributes based on self.mutable flag

        Parameters
        ----------
        name : str
            Attribute name to be assigned to Meta
        value : str or boolean
            String to be assigned to attribute specified by name or boolean
            if name is 'mutable'

        Note
        ----
        @properties are assumed to be mutable.

        We avoid recursively setting properties using
        method from https://stackoverflow.com/a/15751135

        """

        # mutable handled explicitly to avoid recursion
        if name != 'mutable':

            # check if this attribute is a property
            propobj = getattr(self.__class__, name, None)
            if isinstance(propobj, property):
                # check if the property is settable
                if propobj.fset is None:
                    raise AttributeError(''.join("can't set attribute  ",
                                                 name, " to ", value, ", ",
                                                 "property has no fset"))

                # make mutable in case fset needs it to be
                mutable_tmp = self.mutable
                self.mutable = True

                # set the property
                propobj.fset(self, value)

                # restore mutability flag
                self.mutable = mutable_tmp
            else:
                # a normal attribute
                if self.mutable:
                    # use Object to avoid recursion
                    super(Meta, self).__setattr__(name, value)
                else:
                    estr = ' '.join(("Cannot set attribute", name, "to {val!s}",
                                     "since the Meta object attributes are",
                                     "set to immutable.")).format(val=value)
                    raise AttributeError(estr)
        else:
            super(Meta, self).__setattr__(name, value)

    def __setitem__(self, data_vars, input_data):
        """Convenience method for adding metadata.

        Parameters
        ----------
        data_vars : str, list
            Data variable names for the input metadata
        input_data : dict, pds.Series, or Meta
            Input metadata to be assigned

        """

        input_data = deepcopy(input_data)

        if isinstance(input_data, dict):
            # If not passed an iterable, make it one
            if isinstance(data_vars, str):
                data_vars = [data_vars]
                for key in input_data:
                    input_data[key] = [input_data[key]]
            elif isinstance(data_vars, slice) and (data_vars.step is None):
                # Check for use of instrument[indx, :] or instrument[idx]
                data_vars = [dkey for dkey in self.data.keys()]

            # Make sure the variable names are in good shape.  The Meta object
            # is case insensitive, but case preserving. Convert given data_vars
            # into ones Meta has already seen. If new, then input names
            # become the standard
            data_vars = [self.var_case_name(var) for var in data_vars]
            for var in data_vars:
                if var not in self:
                    self._insert_default_values(var)

            # Check if input dict empty.  If so, no metadata was assigned by
            # the user.  This is an empty call and we can head out,
            # since defaults have been assigned
            if len(input_data.keys()) == 0:
                return

            # Perform some checks on the data and make sure number of inputs
            # matches number of metadata inputs.
            for dkey in input_data:
                if len(data_vars) != len(input_data[dkey]):
                    raise ValueError(''.join(('Length of data_vars and inputs',
                                              ' must be equal.')))

            # Make sure the attribute names are in good shape.  Check the
            # attribute's name against existing attribute names.  If the
            # attribute name exists somewhere, then the case of the existing
            # attribute will be enforced upon new data by default for
            # consistency.
            input_keys = [ikey for ikey in input_data]
            for iname in input_keys:
                new_name = self.attr_case_name(iname)
                if new_name != iname:
                    input_data[new_name] = input_data.pop(iname)

            # Time to actually add the metadata
            for ikey in input_data:
                if ikey not in ['children', 'meta']:
                    for i, var in enumerate(data_vars):
                        to_be_set = input_data[ikey][i]
                        if hasattr(to_be_set, '__iter__') \
                           and not isinstance(to_be_set, str):
                            # We have some list-like object that can only
                            # store a single element
                            if len(to_be_set) == 0:
                                # Empty list, ensure there is something to set
                                to_be_set = ['']
                            if isinstance(to_be_set[0], str) \
                                    or isinstance(to_be_set, bytes):
                                if isinstance(to_be_set, bytes):
                                    to_be_set = to_be_set.decode("utf-8")

                                self._data.loc[var, ikey] = '\n\n'.join(
                                    to_be_set)
                            else:
                                warnings.warn(' '.join(('Array elements are',
                                                        'not allowed in meta.',
                                                        'Dropping input :',
                                                        key)))
                        else:
                            self._data.loc[var, ikey] = to_be_set
                else:
                    # key is 'meta' or 'children'
                    # process higher order stuff. Meta inputs could be part of
                    # larger multiple parameter assignment
                    # so not all names may actually have 'meta' to add
                    for j, (item, val) in enumerate(zip(data_vars,
                                                        input_data['meta'])):
                        if val is not None:
                            # Assign meta data, using a recursive call...
                            # heads to if Meta instance call
                            self[item] = val

        elif isinstance(input_data, pds.Series):
            # Outputs from Meta object are a Series. Thus this takes in input
            # from a Meta object. Set data using standard assignment via a dict
            in_dict = input_data.to_dict()
            if 'children' in in_dict:
                child = in_dict.pop('children')
                if child is not None:
                    # if not child.data.empty:
                    self.ho_data[data_vars] = child

            # Remaining items are simply assigned
            self[data_vars] = in_dict

        elif isinstance(input_data, Meta):
            # Dealing with a higher order data set.
            # data_vars is only a single name here (by choice for support)
            if (data_vars in self._ho_data) and (input_data.empty):
                # No actual metadata provided and there is already some
                # higher order metadata in self
                return

            # Get Meta approved variable data_vars
            new_item_name = self.var_case_name(data_vars)

            # Ensure that Meta labels of object to be assigned are
            # consistent with self.  input_data accepts self's labels
            input_data.accept_default_labels(self)

            # Go through and ensure Meta object to be added has variable and
            # attribute names consistent with other variables and attributes
            # this covers custom attributes not handled by default routine
            # above
            attr_names = input_data.attrs()
            new_names = []
            for name in attr_names:
                new_names.append(self.attr_case_name(name))
            input_data.data.columns = new_names

            # Same thing for variables
            var_names = input_data.data.index
            new_names = []
            for name in var_names:
                new_names.append(self.var_case_name(name))
            input_data.data.index = new_names

            # Assign Meta object now that things are consistent with Meta
            # object settings, but first make sure there are lower dimension
            # metadata parameters, passing in an empty dict fills in defaults
            # if there is no existing metadata info
            self[new_item_name] = {}

            # Now add to higher order data
            self._ho_data[new_item_name] = input_data
        return

    def __getitem__(self, key):
        """Convenience method for obtaining metadata.

        Maps to pandas DataFrame.loc method.

        Parameters
        ----------
        key : str, tuple, or list
            A single variable name, a tuple, or a list

        Raises
        ------
        KeyError
            If a properly formatted key is not present
        NotImplementedError
            If the input is not one of the allowed data types

        Examples
        --------
        ::

            meta['name']
            meta['name1', 'units']
            meta[['name1', 'name2'], 'units']
            meta[:, 'units']

            # for higher order data
            meta['name1', 'subvar', 'units']
            meta['name1', ('units', 'scale')]

        """
        # Define a local convenience function
        def match_name(func, var_name, index_or_column):
            """Applies func on input variables(s) depending on variable type
            """
            if isinstance(var_name, str):
                # If variable is a string, use it as input
                return func(var_name)
            elif isinstance(var_name, slice):
                # If variable is a slice, use it to select data from the
                # supplied index or column input
                return [func(var) for var in index_or_column[var_name]]
            else:
                # Otherwise, assume the variable iterable input
                return [func(var) for var in var_name]

        # Access desired metadata based on key data type
        if isinstance(key, tuple):
            # If key is a tuple, looking at index, column access pattern
            if len(key) == 2:
                # If tuple length is 2, index, column
                new_index = match_name(self.var_case_name, key[0],
                                       self.data.index)
                new_name = match_name(self.attr_case_name, key[1],
                                      self.data.columns)
                return self.data.loc[new_index, new_name]

            elif len(key) == 3:
                # If tuple length is 3, index, child_index, column
                new_index = self.var_case_name(key[0])
                new_child_index = self.var_case_name(key[1])
                new_name = self.attr_case_name(key[2])
                return self.ho_data[new_index].data.loc[new_child_index,
                                                        new_name]

        elif isinstance(key, list):
            # If key is a list, selection works as-is
            return self[key, :]

        elif isinstance(key, str):
            # If key is a string, treatment varies based on metadata dimension
            if key in self:
                # Get case preserved string for variable name
                new_key = self.var_case_name(key)

                # Don't need to check if in lower, all variables are always in
                # the lower metadata.
                #
                # Assign meta_row using copy to avoid pandas
                # SettingWithCopyWarning, as suggested in
                # https://www.dataquest.io/blog/settingwithcopywarning/
                meta_row = self.data.loc[new_key].copy()
                if new_key in self.keys_nD():
                    meta_row.at['children'] = self.ho_data[new_key].copy()
                else:
                    meta_row.at['children'] = None  # Return empty meta instance

                return meta_row
            else:
                raise KeyError('Key not found in MetaData')
        else:
            raise NotImplementedError("".join(["No way to handle MetaData key ",
                                               "{}; ".format(key.__repr__()),
                                               "expected tuple, list, or str"]))

    def __contains__(self, data_var):
        """case insensitive check for variable name

        Parameters
        ----------
        data_var : str
            Variable name to check if present within the Meta object.

        Returns
        -------
        does_contain : boolean
            True if input Meta class contains the default labels, False if it
            does not

        """
        does_contain = False

        if data_var.lower() in [ikey.lower() for ikey in self.keys()]:
            does_contain = True

        if not does_contain:
            if data_var.lower() in [ikey.lower() for ikey in self.keys_nD()]:
                does_contain = True

        return does_contain

    def __eq__(self, other_meta):
        """ Check equality between Meta instances

        Parameters
        ----------
        other_meta : Meta
            A second Meta class object

        Returns
        -------
        bool
            True if equal, False if not equal

        Note
        ----
        Good for testing.

        Checks if variable names, attribute names, and metadata values
        are all equal between to Meta objects. Note that this comparison
        treats np.NaN == np.NaN as True.

        Name comparison is case-sensitive.

        """

        if not isinstance(other_meta, Meta):
            # The object being compared wasn't even the correct class
            return NotImplemented

        # Check if the variables and attributes are the same
        for iter1, iter2 in [(self.keys(), other_meta.keys()),
                             (self.attrs(), other_meta.attrs())]:
            list1 = [value for value in iter1]
            list2 = [value for value in iter2]

            try:
                testing.assert_lists_equal(list1, list2)
            except AssertionError:
                return False

        # Check that the values of all elements are the same. NaN is treated
        # as equal, though mathematically NaN is not equal to anything
        for key in self.keys():
            for attr in self.attrs():
                if not testing.nan_equal(self[key, attr],
                                         other_meta[key, attr]):
                    return False

        # Check the higher order products. Recursive call into this function
        # didn't work, so spell out the details.
        keys1 = [key for key in self.keys_nD()]
        keys2 = [key for key in other_meta.keys_nD()]
        try:
            testing.assert_lists_equal(keys1, keys2)
        except AssertionError:
            return False

        # Check the higher order variables within each nD key are the same.
        # NaN is treated as equal, though mathematically NaN is not equal
        # to anything
        for key in self.keys_nD():
            for iter1, iter2 in [(self[key].children.keys(),
                                  other_meta[key].children.keys()),
                                 (self[key].children.attrs(),
                                  other_meta[key].children.attrs())]:
                list1 = [value for value in iter1]
                list2 = [value for value in iter2]

                try:
                    testing.assert_lists_equal(list1, list2)
                except AssertionError:
                    return False

            # Check if all elements are individually equal
            for ckey in self[key].children.keys():
                for cattr in self[key].children.attrs():
                    if not testing.nan_equal(
                            self[key].children[ckey, cattr],
                            other_meta[key].children[ckey, cattr]):
                        return False

        # If we made it this far, things are good
        return True

    # -----------------------------------------------------------------------
    # Define the hidden methods

    def _insert_default_values(self, data_var):
        """Set the default label values for a data variable

        Parameters
        ----------
        data_var : str
            Name of the data variable

        Note
        ----
        Sets NaN for all float values, -1 for all int values, 'data_var' for
        names labels, '' for all other str values, and None for any other
        data type.

        """
        # Cycle through each label type to create a list off label names
        # and label default values
        labels = list()
        default_vals = list()
        for lattr in self.labels.label_type.keys():
            labels.append(getattr(self.labels, lattr))

            if lattr in ['name']:
                default_vals.append(data_var)
            else:
                default_vals.append(self.labels.default_values_from_attr(lattr))

        # Assign the default values to the DataFrame for this data variable
        self._data.loc[data_var, labels] = default_vals

        return

    def _label_setter(self, new_label, current_label, default_type,
                      use_names_default=False):
        """Generalized setter of default meta attributes

        Parameters
        ----------
        new_label : str
            New label to use in the Meta object
        current_label : str
            The current label used within Meta object
        default_type : type
            Type of value to be stored
        use_names_default : bool
            if True, MetaData variable names are used as the default
            value for the specified Meta attributes settings (default=False)

        Note
        ----
        Not intended for end user

        """

        self_attrs = list(self.attrs())
        if new_label not in self_attrs:
            # New label not in metadata
            if current_label in self_attrs:
                # Current label exists and has expected case
                self.data.loc[:, new_label] = self.data.loc[:, current_label]
                self.data = self.data.drop(current_label, axis=1)
            else:
                if self.hasattr_case_neutral(current_label):
                    # There is a similar label with different capitalization
                    current_label = self.attr_case_name(current_label)
                    self.data.loc[:, new_label] = self.data.loc[:,
                                                                current_label]
                    self.data = self.data.drop(current_label, axis=1)
                else:
                    # There is no existing label, setting for the first time
                    if use_names_default:
                        self.data[new_label] = self.data.index
                    else:
                        default_val = self.labels.default_values_from_type(
                            default_type)
                        self.data[new_label] = default_val

            # Check higher order structures and recursively change labels
            for key in self.keys_nD():
                # Update children
                self.ho_data[key]._label_setter(new_label, current_label,
                                                default_type, use_names_default)

        return

    # -----------------------------------------------------------------------
    # Define the public methods and properties

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new_frame):
        """ Set the data property

        Paramters
        ---------
        new_frame : pds.DataFrame
            Data frame containing the metadata, with label names as columns

        """
        self._data = new_frame

    @property
    def ho_data(self):
        return self._ho_data

    @ho_data.setter
    def ho_data(self, new_dict):
        """ Set the higher order data property

        Paramters
        ---------
        new_dict : dict
            Dict containing the higher order data

        """
        self._ho_data = new_dict

    @property
    def empty(self):
        """Return boolean True if there is no metadata

        Returns
        -------
        bool
            Returns True if there is no data, and False if there is data

        """

        # only need to check on lower data since lower data
        # is set when higher metadata assigned
        if self.data.empty:
            return True
        else:
            return False

    def merge(self, other):
        """Adds metadata variables to self that are in other but not in self.

        Parameters
        ----------
        other : pysat.Meta

        """

        for key in other.keys():
            if key not in self:
                # Copies over both lower and higher dimensional data
                self[key] = other[key]
        return

    def drop(self, names):
        """Drops variables (names) from metadata.

        Parameters
        ----------
        names : list-like
            List of string specifying the variable names to drop

        """

        # Drop the lower dimension data
        self.data = self._data.drop(names, axis=0)

        # Drop the higher dimension data
        for name in names:
            if name in self._ho_data:
                self._ho_data.pop(name)
        return

    def keep(self, keep_names):
        """Keeps variables (keep_names) while dropping other parameters

        Parameters
        ----------
        keep_names : list-like
            variables to keep

        """
        # Create a list of variable names to keep
        keep_names = [self.var_case_name(name) for name in keep_names]

        # Get a list of current variable names
        current_names = self._data.index

        # Build a list of variable names to drop
        drop_names = [name for name in current_names if name not in keep_names]

        # Drop names not specified in keep_names list
        self.drop(drop_names)
        return

    def apply_meta_labels(self, other_meta):
        """Applies the existing meta labels from self onto different MetaData

        Parameters
        ----------
        other_meta : Meta
            Meta object to have default labels applied

        Returns
        -------
        other_updated : Meta
            Meta object with the default labels applied

        """
        # Create a copy of other, to avoid altering in place
        other_updated = other_meta.copy()

        # Update the Meta labels
        other_updated.accept_default_labels(self.labels)

        # Return the updated Meta class object
        return other_updated

    def accept_default_labels(self, other_meta):
        """Applies labels for default meta labels from other onto self.

        Parameters
        ----------
        other_meta : Meta
            Meta object to take default labels from

        """

        # Update labels in metadata
        for key in other_meta.labels.label_type:
            new_name = getattr(other_meta.labels, key)
            old_name = getattr(self.labels, key)
            if old_name != new_name:
                self._label_setter(new_name, old_name,
                                   other_meta.labels.label_type[key],
                                   use_names_default=True)

        self.labels = other_meta.labels

        return

    def var_case_name(self, name):
        """Provides stored name (case preserved) for case insensitive input

        Parameters
        ----------
        name : str
            variable name in any case

        Returns
        -------
        out_name : str
            String with case preserved as in the meta object

        Note
        ----
        If name is not found (case-insensitive check) then name is returned,
        as input. This function is intended to be used to help ensure the
        case of a given variable name is the same across the Meta object.

        """

        # Get a lower-case version of the name
        lower_name = name.lower()

        # Cycle through all places where this variable name could be, returning
        # the variable name whose lower-case version matches the lower-case
        # version of the variable name supplied.
        if name in self:
            for out_name in self.keys():
                if lower_name == out_name.lower():
                    return out_name

            for out_name in self.keys_nD():
                if lower_name == out_name.lower():
                    return out_name
        else:
            out_name = name

        return out_name

    def keys(self):
        """Yields variable names stored for 1D variables"""

        for ikey in self.data.index:
            yield ikey

    def keys_nD(self):
        """Yields keys for higher order metadata"""

        for ndkey in self.ho_data:
            yield ndkey

    def attrs(self):
        """Yields metadata products stored for each variable name"""

        for dcol in self.data.columns:
            yield dcol

    def hasattr_case_neutral(self, attr_name):
        """Case-insensitive check for attribute names in this class

        Parameters
        ----------
        attr_name : str
            Name of attribute to find

        Returns
        -------
        has_name : bool
            True if case-insesitive check for attribute name is True

        Note
        ----
        Does not check higher order meta objects

        """
        has_name = False

        if attr_name.lower() in [dcol.lower() for dcol in self.data.columns]:
            has_name = True

        return has_name

    def attr_case_name(self, name):
        """Returns preserved case name for case insensitive value of name.

        Parameters
        ----------
        name : str
            Name of variable to get stored case form

        Returns
        -------
        out_name : str
            Name in proper case

        Note
        ----
        Checks first within standard attributes. If not found there, checks
        attributes for higher order data structures. If not found, returns
        supplied name as it is available for use. Intended to be used to help
        ensure that the same case is applied to all repetitions of a given
        variable name.

        """
        lower_name = name.lower()
        for out_name in self.attrs():
            if lower_name == out_name.lower():
                return out_name

        # check if attribute present in higher order structures
        for key in self.keys_nD():
            for out_name in self[key].children.attrs():
                if lower_name == out_name.lower():
                    return out_name

        # nothing was found if still here
        # pass name back, free to be whatever
        return name

    def concat(self, other_meta, strict=False):
        """Concats two metadata objects together.

        Parameters
        ----------
        other_meta : Meta
            Meta object to be concatenated
        strict : bool
            If True, this flag ensures there are no duplicate variable names
            (default=False)

        Returns
        -------
        mdata : Meta
            Concatenated object

        Note
        ----
        Uses units and name label of self if other_meta is different

        """

        mdata = self.copy()

        # Check the inputs
        if strict:
            for key in other_meta.keys():
                if key in mdata:
                    raise RuntimeError(''.join(('Duplicated keys (variable ',
                                                'names) across Meta ',
                                                'objects in keys().')))
            for key in other_meta.keys_nD():
                if key in mdata:
                    raise RuntimeError(''.join(('Duplicated keys (variable ',
                                                'names) across Meta '
                                                'objects in keys_nD().')))

        # Make sure labels between the two objects are the same
        other_meta_updated = other_meta.copy()
        other_meta_updated.labels = self.labels

        # Concat 1D metadata in data frames to copy of current metadata
        for key in other_meta_updated.keys():
            mdata.data.loc[key] = other_meta.data.loc[key]

        # Combine the higher order meta data
        for key in other_meta_updated.keys_nD():
            mdata.ho_data[key] = other_meta.ho_data[key]

        return mdata

    def copy(self):
        """Deep copy of the meta object."""
        return deepcopy(self)

    def pop(self, label_name):
        """Remove and return metadata about variable

        Parameters
        ----------
        label_name : str
            Meta key for a data variable

        Returns
        -------
        output : pds.Series
            Series of metadata for variable

        """
        # Check if the specified label name is present
        if label_name in self:
            # Get case preserved name for variable
            new_name = self.var_case_name(label_name)

            # Check if the label name is for 1D or nD meta data
            if new_name in self.keys():
                output = self[new_name]
                self.data = self.data.drop(new_name, axis=0)
            else:
                output = self.ho_data.pop(new_name)
        else:
            raise KeyError('Key not present in metadata variables')

        return output

    def transfer_attributes_to_instrument(self, inst, strict_names=False):
        """Transfer non-standard attributes in Meta to Instrument object.

        Parameters
        ----------
        inst : pysat.Instrument
            Instrument object to transfer attributes to
        strict_names : bool
            If True, produces an error if the Instrument object already
            has an attribute with the same name to be copied (default=False).

        Note
        ----
        pysat's load_netCDF and similar routines are only able to attach
        netCDF4 attributes to a Meta object. This routine identifies these
        attributes and removes them from the Meta object. Intent is to
        support simple transfers to the pysat.Instrument object.

        Will not transfer names that conflict with pysat default attributes.

        """

        # Save the base Instrument attributes
        banned = inst._base_attr

        # Current attributes
        inst_attr = dir(inst)

        # Get base attribute set, and attributes attached to instance
        base_attrb = self._base_attr
        this_attrb = dir(self)

        # Collect these attributes into a dict
        adict = {}
        transfer_key = []
        for key in this_attrb:
            if key not in banned:
                if key not in base_attrb:
                    # Don't store any hidden attributes
                    if key[0] != '_':
                        adict[key] = getattr(self, key)
                        transfer_key.append(key)
                        # remove key from meta
                        delattr(self, key)

        # Store any non-standard attributes in Instrument get list of
        # instrument objects attributes first to check if a duplicate
        # instrument attributes stay with instrument
        inst_attr = dir(inst)

        for key in transfer_key:
            # Note, keys in transfer_key already checked against banned
            if key not in inst_attr:
                setattr(inst, key, adict[key])
            else:
                if not strict_names:
                    setattr(inst, key, adict[key])
                else:
                    if not strict_names:
                        # Use naming convention: new_name = 'pysat_attr_' + key
                        inst.__setattr__(key, adict[key])
                    else:
                        rerr = ''.join(('Attribute ', key, 'attached to the '
                                        'Meta object can not be transferred ',
                                        'as it already exists in the ',
                                        'Instrument object.'))
                        raise RuntimeError(rerr)
        return

    @classmethod
    def from_csv(cls, filename=None, col_names=None, sep=None, **kwargs):
        """Create instrument metadata object from csv.

        Parameters
        ----------
        filename : string
            absolute filename for csv file or name of file stored in pandas
            instruments location
        col_names : list-like collection of strings
            column names in csv and resultant meta object
        sep : string
            column seperator for supplied csv filename
        **kwargs : dict
            Optional kwargs used by pds.read_csv

        Note
        ----
        column names must include at least ['name', 'long_name', 'units'],
        assumed if col_names is None.

        """
        req_names = ['name', 'long_name', 'units']
        if col_names is None:
            col_names = req_names
        elif not all([rname in col_names for rname in req_names]):
            raise ValueError('col_names must include name, long_name, units.')

        if sep is None:
            sep = ','

        if filename is None:
            raise ValueError('Must supply an instrument module or file path.')
        elif not isinstance(filename, str):
            raise ValueError('Keyword name must be related to a string')
        elif not os.path.isfile(filename):
            # Not a real file, assume input is a pysat instrument name
            # and look in the standard pysat location.
            testfile = os.path.join(pysat.__path__[0], 'instruments', filename)
            if os.path.isfile(testfile):
                filename = testfile
            else:
                # Try to form an absolute path, if the relative path failed
                testfile = os.path.abspath(filename)
                if not os.path.isfile(testfile):
                    raise ValueError("Unable to create valid file path.")
                else:
                    filename = testfile

        mdata = pds.read_csv(filename, names=col_names, sep=sep, **kwargs)

        if not mdata.empty:
            # Make sure the data name is the index
            mdata.index = mdata['name']
            del mdata['name']
            return cls(metadata=mdata)
        else:
            raise ValueError(''.join(['Unable to retrieve information from ',
                                      filename]))

    # TODO
    # @classmethod
    # def from_nc():
    #     """not implemented yet, load metadata from netCDF"""
    #     pass
    #
    # @classmethod
    # def from_dict():
    #     """not implemented yet, load metadata from dict of items/list types
    #     """
    #     pass


class MetaLabels(object):
    """ Stores metadata labels for Instrument instance

    Parameters
    ----------
    units : tuple
        Units label name and value type (default=('units', str))
    name : tuple
        Name label name and value type (default=('long_name', str))
    notes : tuple
        Notes label name and value type (default=('notes', str))
    desc : tuple
        Description label name and value type (default=('desc', str))
    min_val : tuple
        Minimum value label name and value type (default=('value_min', float))
    max_val : tuple
        Maximum value label name and value type (default=('value_max', float))
    fill_val : tuple
        Fill value label name and value type (default=('fill', float))
    kwargs : dict
        Dictionary containing optional label attributes, where the keys are the
        attribute names and the values are tuples containing the label name and
        value type

    Attributes
    ----------
    data : pandas.DataFrame
        index is variable standard name, 'units', 'long_name', and other
        defaults are also stored along with additional user provided labels.
    units : str
        String used to label units in storage. (default='units')
    name : str
        String used to label long_name in storage. (default='long_name')
    notes : str
       String used to label 'notes' in storage. (default='notes')
    desc : str
       String used to label variable descriptions in storage.
       (default='desc')
    min_val : str
       String used to label typical variable value min limit in storage.
       (default='value_min')
    max_val : str
       String used to label typical variable value max limit in storage.
       (default='value_max')
    fill_val : str
        String used to label fill value in storage. The default follows the
        netCDF4 standards (default='fill')

    Note
    ----
    Meta object preserves the case of variables and attributes as it first
    receives the data. Subsequent calls to set new metadata with the same
    variable or attribute will use case of first call. Accessing or setting
    data thereafter is case insensitive. In practice, use is case insensitive
    but the original case is preserved. Case preseveration is built in to
    support writing files with a desired case to meet standards.

    Metadata for higher order data objects, those that have
    multiple products under a single variable name in a pysat.Instrument
    object, are stored by providing a Meta object under the single name.

    Supports any custom metadata values in addition to the expected metadata
    attributes (units, name, notes, desc, value_min, value_max, and fill).
    These base attributes may be used to programatically access and set types
    of metadata regardless of the string values used for the attribute. String
    values for attributes may need to be changed depending upon the standards
    of code or files interacting with pysat.

    Meta objects returned as part of pysat loading routines are automatically
    updated to use the same values of units, etc. as found in the
    pysat.Instrument object.

    """

    def __init__(self, metadata=None, units=('units', str),
                 name=('long_name', str), notes=('notes', str),
                 desc=('desc', str),
                 min_val=('value_min', float), max_val=('value_max', float),
                 fill_val=('fill', float), **kwargs):
        """ Initialize the MetaLabels class

        Parameters
        ----------
        units : tuple
            Units label name and value type (default=('units', str))
        name : tuple
            Name label name and value type (default=('long_name', str))
        notes : tuple
            Notes label name and value type (default=('notes', str))
        desc : tuple
            Description label name and value type (default=('desc', str))
        min_val : tuple
            Minimum value label name and value type
            (default=('value_min', float))
        max_val : tuple
            Maximum value label name and value type
            (default=('value_max', float))
        fill_val : tuple
            Fill value label name and value type (default=('fill', float))
        kwargs : dict
            Dictionary containing optional label attributes, where the keys
            are the attribute names and the values are tuples containing the
            label name and value type

        """
        # Initialize the coupled metadata
        self.meta = metadata

        # Initialize a dictionary of label types, whose keys are the label
        # attributes
        self.label_type = {'units': units[1], 'name': name[1],
                           'notes': notes[1], 'desc': desc[1],
                           'min_val': min_val[1], 'max_val': max_val[1],
                           'fill_val': fill_val[1]}

        # Set the default labels and types
        self.units = units[0]
        self.name = name[0]
        self.notes = notes[0]
        self.desc = desc[0]
        self.min_val = min_val[0]
        self.max_val = max_val[0]
        self.fill_val = fill_val[0]

        # Set the custom labels and label types
        for custom_label in kwargs.keys():
            setattr(self, custom_label, kwargs[custom_label][0])
            self.label_type[custom_label] = kwargs[custom_label][1]

        return

    def __setattr__(self, name, value):
        """Conditionally sets attributes based on their type

        Parameters
        ----------
        name : str
            Attribute name to be assigned to MetaLabels
        value
            Value (any type) to be assigned to attribute specified by name

        """
        # Get old attribute value for reference
        if hasattr(self, name):
            old_value = getattr(self, name)
        else:
            old_value = None

        # Use Object to avoid recursion
        super(MetaLabels, self).__setattr__(name, value)

        # Before setting the attribute, see if upstream changes are needed
        if old_value is not None and name not in ['label_type', 'meta']:
            if hasattr(self, 'meta') and hasattr(self.meta, 'data'):
                self.meta._label_setter(value, getattr(self, name),
                                        self.label_type[name],
                                        use_names_default=True)

    def __repr__(self):
        """String describing MetaData instantiation parameters

        Returns
        -------
        out_str : str
            Simply formatted output string

        """
        label_str = ', '.join(["{:s}={:} {:}".format(mlab, getattr(self, mlab),
                                                     self.label_type[mlab])
                               for mlab in self.label_type.keys()])
        out_str = ''.join(['pysat.MetaLabels(', label_str, ")"])
        return out_str

    def __str__(self):
        """String describing Meta instance, variables, and attributes

        Returns
        -------
        out_str : str
            Nicely formatted output string

        """
        # Set the printing limits and get the label attributes
        ncol = 3
        lab_attrs = ["{:s}->{:s}".format(mlab, getattr(self, mlab))
                     for mlab in self.label_type.keys()]
        nlabels = len(lab_attrs)

        # Print the MetaLabels
        out_str = "MetaLabels:\n"
        out_str += "-----------\n"
        out_str += core_utils.fmt_output_in_cols(lab_attrs, ncols=ncol,
                                                 max_num=nlabels)

        return out_str

    def default_values_from_type(self, val_type):
        """ Return the default values for each label based on their type

        Parameters
        ----------
        val_type : type
            Variable type for the value to be assigned to a MetaLabel

        Returns
        -------
        default_val : str, float, int, NoneType
            Sets NaN for all float values, -1 for all int values, and '' for
            all str values except for 'scale', which defaults to 'linear', and
            None for any othere data type

        """

        # Assign the default value
        if issubclass(val_type, str):
            default_val = ''
        elif val_type is float:
            default_val = np.nan
        elif val_type is int:
            default_val = -1
        else:
            default_val = None

        return default_val

    def default_values_from_attr(self, attr_name):
        """ Return the default values for each label based on their type

        Parameters
        ----------
        attr_name : str
            Label attribute name (e.g., max_val)

        Returns
        -------
        default_val : str, float, int, NoneType
            Sets NaN for all float values, -1 for all int values, and '' for
            all str values except for 'scale', which defaults to 'linear', and
            None for any othere data type

        Raises
        ------
        ValueError
            For unknown attr_name

        """

        # Test the input parameter
        if attr_name not in self.label_type.keys():
            raise ValueError('unknown label attribute {:}'.format(attr_name))

        # Assign the default value
        if attr_name == 'scale':
            default_val = 'linear'
        else:
            default_val = self.default_values_from_type(
                self.label_type[attr_name])

        return default_val



"""
tests the pysat utils.time area
"""
import numpy as np

from nose.tools import raises
import pandas as pds

from pysat.utils import time as pytime


###########
# getyrdoy

def test_getyrdoy_1():
    """Test the date to year, day of year code functionality"""

    date = pds.datetime(2009, 1, 1)
    yr, doy = pytime.getyrdoy(date)

    assert ((yr == 2009) & (doy == 1))


def test_getyrdoy_leap_year():
    """Test the date to year, day of year code functionality (leap_year)"""

    date = pds.datetime(2008, 12, 31)
    yr, doy = pytime.getyrdoy(date)

    assert ((yr == 2008) & (doy == 366))


#############
# parse_date

def test_parse_date_2_digit_year():
    """Test the ability to parse a str to produce a pandas datetime"""

    date = pytime.parse_date('14', '10', '31')

    assert date == pds.datetime(2014, 10, 31)


def test_parse_date_2_digit_year_last_century():
    """Test the ability to parse a str to produce a pandas datetime
    pre-2000"""

    date = pytime.parse_date('94', '10', '31', century=1900)

    assert date == pds.datetime(1994, 10, 31)


def test_parse_date_4_digit_year():
    """Test the ability to parse a str to produce a pandas datetime"""

    date = pytime.parse_date('1994', '10', '31')

    assert date == pds.datetime(1994, 10, 31)


@raises(ValueError)
def test_parse_date_bad_input():
    """Test the ability to idenitfy a non-physical date"""

    _ = pytime.parse_date('194', '15', '31')


############
# calc_freq

def test_calc_freq():
    """Test index frequency calculation"""

    tind = pytime.create_datetime_index(year=np.ones(shape=(4,))*2001,
                                        month=np.ones(shape=(4,)),
                                        uts=np.arange(0.0, 4.0, 1.0))
    freq = pytime.calc_freq(tind)

    assert freq.find("1S") == 0


def test_calc_freq_ns():
    """Test index frequency calculation with nanosecond output"""

    tind = pytime.create_datetime_index(year=np.ones(shape=(4,))*2001,
                                        month=np.ones(shape=(4,)),
                                        uts=np.arange(0.0, 0.04, .01))
    freq = pytime.calc_freq(tind)

    assert freq.find("10000000N") == 0


@raises(ValueError)
def test_calc_freq_len_fail():
    """Test index frequency calculation with empty list"""

    pytime.calc_freq(list())


@raises(AttributeError)
def test_calc_freq_type_fail():
    """Test index frequency calculation with non-datetime list"""

    pytime.calc_freq([1, 2, 3, 4])


####################
# create_date_range

def test_create_date_range():
    """Test ability to generate season list"""

    start = pds.datetime(2012, 2, 28)
    stop = pds.datetime(2012, 3, 1)
    season = pytime.create_date_range(start, stop, freq='D')

    assert season[0] == start
    assert season[-1] == stop
    assert len(season) == 3


def test_create_date_range_w_gaps():
    """Test ability to generate season list"""

    start = [pds.datetime(2012, 2, 28), pds.datetime(2013, 2, 28)]
    stop = [pds.datetime(2012, 3, 1), pds.datetime(2013, 3, 1)]
    season = pytime.create_date_range(start, stop, freq='D')

    assert season[0] == start[0]
    assert season[-1] == stop[-1]
    assert len(season) == 5


#########################
# create_datetime_index

def test_create_datetime_index():
    """Tests ability to create an array of datetime objects from distinct
    arrays of input paramters"""

    arr = np.ones(4)

    dates = pytime.create_datetime_index(year=2012*arr, month=2*arr,
                                         day=28*arr, uts=np.arange(0, 4))

    assert dates[0] == pds.datetime(2012, 2, 28)
    assert dates[-1] == pds.datetime(2012, 2, 28, 0, 0, 3)
    assert len(dates) == 4


@raises(ValueError)
def test_create_datetime_index_wo_year():
    """Must include a year"""

    _ = pytime.create_datetime_index()


def test_create_datetime_index_wo_month_day_uts():
    """Tests ability to generate missing paramters"""

    arr = np.ones(4)

    dates = pytime.create_datetime_index(year=2012*arr)

    assert dates[0] == pds.datetime(2012, 1, 1)
    assert dates[-1] == pds.datetime(2012, 1, 1)
    assert len(dates) == 4


def test_deprecated_season_date_range():
    """Tests that deprecation of season_date_range is working"""

    import warnings

    start = pds.datetime(2012, 2, 28)
    stop = pds.datetime(2012, 3, 1)
    warnings.simplefilter("always")
    with warnings.catch_warnings(record=True) as war1:
        season1 = pytime.create_date_range(start, stop, freq='D')
    with warnings.catch_warnings(record=True) as war2:
        season2 = pytime.season_date_range(start, stop, freq='D')

    assert len(season1) == len(season2)
    assert (season1 == season2).all()
    assert len(war1) == 0
    assert len(war2) == 1
    assert war2[0].category == DeprecationWarning



import datetime as dt
import numpy as np
import os
import pandas as pds
import warnings

import pysat

logger = pysat.logger

ackn_str = ' '.join(("Test instruments provided through the pysat project.",
                     "https://www.github.com/pysat/pysat"))

# Load up citation information
with pysat.utils.NetworkLock(os.path.join(pysat.here, 'citation.txt'), 'r') as \
        locked_file:
    refs = locked_file.read()


def init(self, test_init_kwarg=None):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    Shifts time index of files by 5-minutes if mangle_file_dates
    set to True at pysat.Instrument instantiation.

    Creates a file list for a given range if the file_date_range
    keyword is set at instantiation.

    Parameters
    ----------
    self : pysat.Instrument
        This object
    test_init_kwarg : any or NoneType
        Testing keyword (default=None)

    """

    logger.info(ackn_str)
    self.acknowledgements = ackn_str
    self.references = refs

    # Assign parameters for testing purposes
    self.new_thing = True
    self.test_init_kwarg = test_init_kwarg

    return


def clean(self, test_clean_kwarg=None):
    """Cleaning function

    Parameters
    ----------
    self : pysat.Instrument
        This object
    test_clean_kwarg : any or NoneType
        Testing keyword (default=None)

    """

    self.test_clean_kwarg = test_clean_kwarg

    return


# Optional method
def preprocess(self, test_preprocess_kwarg=None):
    """Customization method that performs standard preprocessing.

    This routine is automatically applied to the Instrument object
    on every load by the pysat nanokernel (first in queue). Object
    modified in place.

    Parameters
    ----------
    test_preprocess_kwarg : any or NoneType
        Testing keyword (default=None)

    """

    self.test_preprocess_kwarg = test_preprocess_kwarg

    return


def list_files(tag=None, inst_id=None, data_path=None, format_str=None,
               file_date_range=None, test_dates=None, mangle_file_dates=False,
               test_list_files_kwarg=None):
    """Produce a fake list of files spanning three years

    Parameters
    ----------
    tag : str or NoneType
        pysat instrument tag (default=None)
    inst_id : str or NoneType
        pysat satellite ID tag (default=None)
    data_path : str or NoneType
        pysat data path (default=None)
    format_str : str or NoneType
        file format string (default=None)
    file_date_range : pds.date_range
        File date range. The default mode generates a list of 3 years of daily
        files (1 year back, 2 years forward) based on the test_dates passed
        through below.  Otherwise, accepts a range of files specified by the
        user.
        (default=None)
    test_dates : dt.datetime or NoneType
        Pass the _test_date object through from the test instrument files
    mangle_file_dates : bool
        If True, file dates are shifted by 5 minutes. (default=False)
    test_list_files_kwarg : any or NoneType
        Testing keyword (default=None)

    Returns
    -------
    Series of filenames indexed by file time

    """

    # Support keyword testing
    logger.info(''.join(('test_list_files_kwarg = ',
                         str(test_list_files_kwarg))))

    if data_path is None:
        data_path = ''

    # Determine the appropriate date range for the fake files
    if file_date_range is None:
        start = test_dates[''][''] - pds.DateOffset(years=1)
        stop = (test_dates[''][''] + pds.DateOffset(years=2)
                - pds.DateOffset(days=1))
        file_date_range = pds.date_range(start, stop)

    index = file_date_range

    # Mess with file dates if kwarg option set
    if mangle_file_dates:
        index = index + dt.timedelta(minutes=5)

    # Create the list of fake filenames
    names = [data_path + date.strftime('%Y-%m-%d') + '.nofile'
             for date in index]

    return pds.Series(names, index=index)


def list_remote_files(tag=None, inst_id=None, data_path=None, format_str=None,
                      start=None, stop=None, test_dates=None, user=None,
                      password=None, mangle_file_dates=False,
                      test_list_remote_kwarg=None):
    """Produce a fake list of files spanning three years and one month to
    simulate new data files on a remote server

    Parameters
    ----------
    tag : str or NoneType
        pysat instrument tag (default=None)
    inst_id : str or NoneType
        pysat satellite ID tag (default=None)
    data_path : str or NoneType
        pysat data path (default=None)
    format_str : str or NoneType
        file format string (default=None)
    start : dt.datetime or NoneType
        Starting time for file list. A None value will start 1 year before
        test_date
        (default=None)
    stop : dt.datetime or NoneType
        Ending time for the file list.  A None value will stop 2 years 1 month
        after test_date
        (default=None)
    test_dates : dt.datetime or NoneType
        Pass the _test_date object through from the test instrument files
    user : string or NoneType
        User string input used for download. Provided by user and passed via
        pysat. If an account is required for dowloads this routine here must
        error if user not supplied. (default=None)
    password : string or NoneType
        Password for data download. (default=None)
    mangle_file_dates : bool
        If True, file dates are shifted by 5 minutes. (default=False)
    test_list_remote_kwarg : any or NoneType
        Testing keyword (default=None)

    Returns
    -------
    pds.Series
        Filenames indexed by file time, see list_files for more info

    """

    # Support keyword testing
    logger.info(''.join(('test_list_remote_kwarg = ',
                         str(test_list_remote_kwarg))))

    # Determine the appropriate date range for the fake files
    if start is None:
        start = test_dates[''][''] - pds.DateOffset(years=1)

    if stop is None:
        stop = (test_dates[''][''] + pds.DateOffset(years=2)
                - pds.DateOffset(days=1) + pds.DateOffset(months=1))

    file_date_range = pds.date_range(start, stop)

    return list_files(tag=tag, inst_id=inst_id, data_path=data_path,
                      format_str=format_str, file_date_range=file_date_range,
                      mangle_file_dates=mangle_file_dates,
                      test_dates=test_dates)


def download(date_array, tag, inst_id, data_path=None, user=None,
             password=None, test_download_kwarg=None):
    """Simple pass function for pysat compatibility for test instruments.

    This routine is invoked by pysat and is not intended for direct use by the
    end user.

    Parameters
    ----------
    date_array : array-like
        list of datetimes to download data for. The sequence of dates need not
        be contiguous.
    tag : string
        Tag identifier used for particular dataset. This input is provided by
        pysat. (default='')
    inst_id : string
        Instrument ID string identifier used for particular dataset. This input
        is provided by pysat. (default='')
    data_path : string or NoneType
        Path to directory to download data to. (default=None)
    user : string or NoneType
        User string input used for download. Provided by user and passed via
        pysat. If an account is required for downloads this routine here must
        error if user not supplied. (default=None)
    password : string or NoneType
        Password for data download. (default=None)
    test_download_kwarg : any or NoneType
        Testing keyword (default=None)

    Raises
    ------
    ValueError
        When user/password are required but not supplied

    Warnings
    --------
    When no download support will be provided

    """

    # Support keyword testing
    logger.info(''.join(('test_download_kwarg = ', str(test_download_kwarg))))

    if tag == 'no_download':
        warnings.warn('This simulates an instrument without download support')

    # Check that user name and password are passed through the unit tests
    if tag == 'user_password':
        if (not user) and (not password):
            # Note that this line will be uncovered if test succeeds
            raise ValueError(' '.join(('Tests are not passing user and',
                                       'password to test instruments')))

    return


def generate_fake_data(t0, num_array, period=5820, data_range=[0.0, 24.0],
                       cyclic=True):
    """Generates fake data over a given range

    Parameters
    ----------
    t0 : float
        Start time in seconds
    num_array : array_like
        Array of time steps from t0.  This is the index of the fake data
    period : int
        The number of seconds per period.
        (default = 5820)
    data_range : float
        For cyclic functions, the range of data values cycled over one period.
        Not used for non-cyclic functions.
        (default = 24.0)
    cyclic : bool
        If True, assume that fake data is a cyclic function (ie, longitude,
        slt) that will reset to data_range[0] once it reaches data_range[1].
        If False, continue to monotonically increase

    Returns
    -------
    data : array-like
        Array with fake data

    """

    if cyclic:
        uts_root = np.mod(t0, period)
        data = (np.mod(uts_root + num_array, period)
                * (np.diff(data_range)[0] / float(period))) + data_range[0]
    else:
        data = ((t0 + num_array) / period).astype(int)

    return data


def generate_times(fnames, num, freq='1S'):
    """Construct list of times for simulated instruments

    Parameters
    ----------
    fnames : list
        List of filenames.
    num : int
        Number of times to generate
    freq : string
        Frequency of temporal output, compatible with pandas.date_range
        [default : '1S']

    Returns
    -------
    uts : array
        Array of integers representing uts for a given day
    index : pds.DatetimeIndex
        The DatetimeIndex to be used in the pysat test instrument objects
    date : datetime
        The requested date reconstructed from the fake file name

    """

    if isinstance(num, str):
        estr = ''.join(('generate_times support for input strings interpreted ',
                        'as the number of times has been deprecated. Please ',
                        'switch to using integers.'))
        warnings.warn(estr, DeprecationWarning)

    uts = []
    indices = []
    dates = []
    for loop, fname in enumerate(fnames):
        # Grab date from filename
        parts = os.path.split(fname)[-1].split('-')
        yr = int(parts[0])
        month = int(parts[1])
        day = int(parts[2][0:2])
        date = dt.datetime(yr, month, day)
        dates.append(date)

        # Create one day of data at desired frequency
        end_date = date + dt.timedelta(seconds=86399)
        index = pds.date_range(start=date, end=end_date, freq=freq)
        index = index[0:num]
        indices.extend(index)
        uts.extend(index.hour * 3600 + index.minute * 60 + index.second
                   + 86400. * loop)

    # Combine index times together
    index = pds.DatetimeIndex(indices)

    # Make UTS an array
    uts = np.array(uts)

    return uts, index, dates


def define_period():
    """Define the default periods for the fake data functions

    Returns
    -------
    def_period : dict
        Dictionary of periods to use in test instruments

    Note
    ----
    Local time and longitude slightly out of sync to simulate motion of Earth

    """

    def_period = {'lt': 5820,  # 97 minutes
                  'lon': 6240,  # 104 minutes
                  'angle': 5820}

    return def_period


def define_range():
    """Define the default ranges for the fake data functions

    Returns
    -------
    def_range : dict
        Dictionary of periods to use in test instruments

    """

    def_range = {'lt': [0.0, 24.0],
                 'lon': [0.0, 360.0],
                 'angle': [0.0, 2.0 * np.pi]}

    return def_range


def eval_dep_warnings(warns, check_msgs):
    """Evaluate deprecation warnings by category and message

    Parameters
    ----------
    warns : list
        List of warnings.WarningMessage objects
    check_msgs : list
        List of strings containing the expected warning messages

    Returns
    -------
    found_msgs : list
        List of booleans corresponding to `check_msgs`, which are True when
        the messages are found and False when they are not

    """

    # Initialize the output
    found_msgs = [False for msg in check_msgs]

    # Test the warning messages, ensuring each attribute is present
    for iwar in warns:
        if iwar.category == DeprecationWarning:
            for i, msg in enumerate(check_msgs):
                if str(iwar.message).find(msg) >= 0:
                    found_msgs[i] = True

    return found_msgs



# -*- coding: utf-8 -*-
"""
Produces fake instrument data for testing.
"""
from __future__ import print_function
from __future__ import absolute_import
import functools
import numpy as np

import pandas as pds

import pysat
from pysat.instruments.methods import testing as mm_test

platform = 'pysat'
name = 'testing2d'
tags = {'': 'Regular testing data set'}
sat_ids = {'': ['']}
_test_dates = {'': {'': pysat.datetime(2009, 1, 1)}}


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    Parameters
    ----------
    self : pysat.Instrument
        This object

    """

    self.new_thing = True


def default(inst):
    """Default customization function.

    This routine is automatically applied to the Instrument object
    on every load by the pysat nanokernel (first in queue).

    Parameters
    ----------
    self : pysat.Instrument
        This object

    """

    pass


def load(fnames, tag=None, sat_id=None, malformed_index=False):
    """ Loads the test files

    Parameters
    ----------
    fnames : list
        List of filenames
    tag : str or NoneType
        Instrument tag (accepts '')
    sat_id : str or NoneType
        Instrument satellite ID (accepts '' or a number (i.e., '10'), which
        specifies the number of data points to include in the test instrument)
    malformed_index : bool
        If True, the time index will be non-unique and non-monotonic.
        (default=False)

    Returns
    -------
    data : pds.DataFrame
        Testing data
    meta : pysat.Meta
        Metadataxs

    """

    # create an artifical satellite data set
    iperiod = mm_test.define_period()
    drange = mm_test.define_range()
    # Using 100s frequency for compatibility with seasonal analysis unit tests
    uts, index, date = mm_test.generate_times(fnames, sat_id, freq='100S')
    # seed DataFrame with UT array
    data = pysat.DataFrame(uts, columns=['uts'])

    # need to create simple orbits here. Have start of first orbit
    # at 2009,1, 0 UT. 14.84 orbits per day
    # figure out how far in time from the root start
    # use that info to create a signal that is continuous from that start
    # going to presume there are 5820 seconds per orbit (97 minute period)
    time_delta = date - pysat.datetime(2009, 1, 1)
    # mlt runs 0-24 each orbit.
    data['mlt'] = mm_test.generate_fake_data(time_delta.total_seconds(), uts,
                                             period=iperiod['lt'],
                                             data_range=drange['lt'])
    # do slt, 20 second offset from mlt
    data['slt'] = mm_test.generate_fake_data(time_delta.total_seconds()+20,
                                             uts, period=iperiod['lt'],
                                             data_range=drange['lt'])
    # create a fake longitude, resets every 6240 seconds
    # sat moves at 360/5820 deg/s, Earth rotates at 360/86400, takes extra time
    # to go around full longitude
    data['longitude'] = mm_test.generate_fake_data(time_delta.total_seconds(),
                                                   uts, period=iperiod['lon'],
                                                   data_range=drange['lon'])
    # create latitude signal for testing polar orbits
    angle = mm_test.generate_fake_data(time_delta.total_seconds(),
                                       uts, period=iperiod['angle'],
                                       data_range=drange['angle'])
    data['latitude'] = 90.0 * np.cos(angle)

    if malformed_index:
        index = index.tolist()
        # nonmonotonic
        index[0:3], index[3:6] = index[3:6], index[0:3]
        # non unique
        index[6:9] = [index[6]]*3

    data.index = index
    data.index.name = 'epoch'
    # higher rate time signal (for scalar >= 2)
    # this time signal used for 2D profiles associated with each time in main
    # DataFrame
    high_rate_template = pds.date_range(date,
                                        date + pds.DateOffset(hours=0,
                                                              minutes=1,
                                                              seconds=39),
                                        freq='2S')

    # create a few simulated profiles
    # DataFrame at each time with mixed variables
    profiles = []
    # DataFrame at each time with numeric variables only
    alt_profiles = []
    # Serie at each time, numeric data only
    series_profiles = []
    # frame indexed by date times
    frame = pds.DataFrame({'density': data.loc[data.index[0:50],
                                               'mlt'].values.copy(),
                           'dummy_str': ['test'] * 50,
                           'dummy_ustr': [u'test'] * 50},
                          index=data.index[0:50],
                          columns=['density', 'dummy_str', 'dummy_ustr'])
    # frame indexed by float
    dd = np.arange(50) * 1.2
    ff = np.arange(50) / 50.
    ii = np.arange(50) * 0.5
    frame_alt = pds.DataFrame({'density': dd, 'fraction': ff},
                              index=ii,
                              columns=['density', 'fraction'])
    # series version of storage
    series_alt = pds.Series(dd, index=ii, name='series_profiles')

    for time in data.index:
        frame.index = high_rate_template + (time - data.index[0])
        profiles.append(frame)
        alt_profiles.append(frame_alt)
        series_profiles.append(series_alt)
    # store multiple data types into main frame
    data['profiles'] = pds.Series(profiles, index=data.index)
    data['alt_profiles'] = pds.Series(alt_profiles, index=data.index)
    data['series_profiles'] = pds.Series(series_profiles, index=data.index)
    return data, meta.copy()


list_files = functools.partial(mm_test.list_files, test_dates=_test_dates)
download = functools.partial(mm_test.download)


# create very limited metadata
meta = pysat.Meta()
meta['uts'] = {'units': 's', 'long_name': 'Universal Time'}
meta['mlt'] = {'units': 'hours', 'long_name': 'Magnetic Local Time'}
meta['slt'] = {'units': 'hours', 'long_name': 'Solar Local Time'}
meta['longitude'] = {'units': 'degrees', 'long_name': 'Longitude'}
meta['latitude'] = {'units': 'degrees', 'long_name': 'Latitude'}
series_profile_meta = pysat.Meta()
series_profile_meta['series_profiles'] = {'units': '', 'long_name': 'series'}
meta['series_profiles'] = {'meta': series_profile_meta, 'units': '',
                           'long_name': 'series'}
profile_meta = pysat.Meta()
profile_meta['density'] = {'units': '', 'long_name': 'profiles'}
profile_meta['dummy_str'] = {'units': '', 'long_name': 'profiles'}
profile_meta['dummy_ustr'] = {'units': '', 'long_name': 'profiles'}
meta['profiles'] = {'meta': profile_meta, 'units': '', 'long_name': 'profiles'}
alt_profile_meta = pysat.Meta()
alt_profile_meta['density'] = {'units': '', 'long_name': 'profiles'}
alt_profile_meta['fraction'] = {'units': '', 'long_name': 'profiles'}
meta['alt_profiles'] = {'meta': alt_profile_meta, 'units': '',
                        'long_name': 'profiles'}



# -*- coding: utf-8 -*-
"""Provides non-instrument routines for DEMETER microsatellite data"""

from __future__ import absolute_import, division, print_function

import numpy as np
import pysat

import logging
logger = logging.getLogger(__name__)


def download(date_array, tag, sat_id, data_path=None, user=None,
             password=None):
    """ Download

    """
    url = 'https://cdpp-archive.cnes.fr/'
    logger.info('Data must be downloaded by registered users at: {:s}'.format(url))
    return


def list_remote_files(tag, sat_id):
    """Lists files available for Demeter.

    Note
    ----
    Unfortunately, obtaining a list of remote files is not
    avilable for this instrument. This function is expected
    by pysat thus it is located here as a placeholder.

    Parameters
    ----------
    tag : (string or NoneType)
        Denotes type of file to load.  Accepted types are <tag strings>.
        (default=None)
    sat_id : (string or NoneType)
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)

    """

    logger.info('Remote file lists are not supported for Demeter.')
    return


def bytes_to_float(chunk):
    """ Convert a chunk of bytes to a float

    Parameters
    ------------
    chunk : string or bytes
       A chunk of bytes

    Returns
    --------
    value : float
        A 32 bit float

    """
    import sys
    import struct
    import codecs

    chunk_code = codecs.encode(chunk, 'hex')

    if sys.version_info.major == 2:
        decoded = chunk_code.decode('hex')
    elif hasattr(chunk_code, "decode"):
        decoded = bytes.fromhex(chunk_code.decode('utf-8'))
    else:
        decoded = bytes.fromhex(chunk_code)

    return struct.unpack("!f", decoded)[0]


def load_general_header(fhandle):
    """ Load the general header block (block 1 for each time)

    Parameters
    ------------
    fhandle : (file handle)
        File handle

    Returns
    ----------
    data : list
        List of data values containing: P field,
        Number of days from 01/01/1950, number of miliseconds in the day,
        UT as datetime, Orbit number, downward (False) upward (True) indicator
    meta : dict
        Dictionary with meta data for keys: 'telemetry station',
        'software processing version', 'software processing subversion',
        'calibration file version', and 'calibration file subversion',
        'data names', 'data units'

    """
    import codecs  # ensures encode is python 2/3 compliant
    import datetime as dt

    block_size = 38  # Position block is 38 bytes
    chunk = fhandle.read(block_size)

    if len(chunk) != block_size:
        return list(), dict()

    # Extract UT timestamp for conversion into datetime object
    year = int(codecs.encode(chunk[8:10], 'hex'), 16)
    month = int(codecs.encode(chunk[10:12], 'hex'), 16)
    day = int(codecs.encode(chunk[12:14], 'hex'), 16)
    hour = int(codecs.encode(chunk[14:16], 'hex'), 16)
    minute = int(codecs.encode(chunk[16:18], 'hex'), 16)
    sec = int(codecs.encode(chunk[18:20], 'hex'), 16)
    microsec = int(codecs.encode(chunk[20:22], 'hex'), 16) * 1000
    dtime = dt.datetime(year, month, day, hour, minute, sec, microsec)

    data = [int(codecs.encode(chunk[:1], 'hex'), 16),  # P field
            int(codecs.encode(chunk[1:4], 'hex'), 16),  # Days since 1/1/1950
            int(codecs.encode(chunk[4:8], 'hex'), 16),  # Millisec in day
            dtime,  # UT timestamp
            int(codecs.encode(chunk[22:24], 'hex'), 16),  # Orbit number
            bool(int(codecs.encode(chunk[24:26], 'hex'), 16))]  # Orbit type

    meta = {'telemetry station': chunk[26:34],
            'software processing version': int(codecs.encode(chunk[34:35],
                                                             'hex'), 16),
            'software processing subversion': int(codecs.encode(chunk[35:36],
                                                                'hex'), 16),
            'calibration file version': int(codecs.encode(chunk[36:37], 'hex'),
                                            16),
            'calibration file subversion': int(codecs.encode(chunk[37:38],
                                                             'hex'), 16),
            'data names': ['P_field', 'epoch_time', 'time_of_day', 'UT',
                           'orbit_number', 'orbit_type'],
            'data units': {'P_field': 'N/A',
                           'epoch_time': 'days since 1/1/1950',
                           'time_of_day': 'ms', 'UT': 'datetime',
                           'orbit_number': 'N/A', 'orbit_type': 'N/A'}}

    return data, meta


def load_location_parameters(fhandle):
    """ Load the orbital and geomagnetic parameter block (block 1 for each
    time)

    Parameters
    ------------
    fhandle : (file handle)
        File handle

    Returns
    ----------
    data : list
        List of data values containing: geoc lat, geoc lon, alt, lt, geom
        lat, geom lon, mlt, inv lat, L-shell, geoc lat of conj point, geoc
        lon of conj point, geoc lat of N conj point at 110 km, geoc lon of
        N conj point at 110 km, geoc lat of S conj point at 110 km, geoc
        lon of S conj point at 110 km, components of magnetic field at sat
        point, proton gyrofreq at sat point, solar position in geog coords
    meta : dict
        Dictionary with meta data for keys: 'software processing version',
        'software processing subversion', 'data names', 'data units'

    """
    import codecs  # ensures encode is python 2/3 compliant

    block_size = 90  # Position block is 90 bytes
    chunk = fhandle.read(block_size)

    if len(chunk) != block_size:
        return list(), dict()

    data = [bytes_to_float(chunk[0:4]),  # geocentric latitude
            bytes_to_float(chunk[4:8]),  # geocentric longitude
            bytes_to_float(chunk[8:12]),  # altitude
            bytes_to_float(chunk[12:16]),  # LT hours of day
            bytes_to_float(chunk[16:20]),  # geomagnetic latitude
            bytes_to_float(chunk[20:24]),  # geomagnetic longitude
            bytes_to_float(chunk[24:28]),  # MLT in hours
            bytes_to_float(chunk[28:32]),  # invariant latitude
            bytes_to_float(chunk[32:36]),  # Mc Ilwain parameter L
            bytes_to_float(chunk[36:40]),  # geocentric latitude of conj point
            bytes_to_float(chunk[40:44]),  # geocentric longitude of conj point
            bytes_to_float(chunk[44:48]),  # geoc lat of N conj point at 110 km
            bytes_to_float(chunk[48:52]),  # geoc lon of N conj point at 110 km
            bytes_to_float(chunk[52:56]),  # geoc lat of S conj point at 110 km
            bytes_to_float(chunk[56:60]),  # geoc lon of S conj point at 110 km
            bytes_to_float(chunk[60:64]),  # mag field component at sat point
            bytes_to_float(chunk[64:68]),  # mag field component at sat point
            bytes_to_float(chunk[68:72]),  # mag field component at sat point
            bytes_to_float(chunk[72:76]),  # Proton gyrofrequency at sat oint
            bytes_to_float(chunk[76:80]),  # Xs in geographic coordinate system
            bytes_to_float(chunk[80:84]),  # Ys in geographic coordinate system
            bytes_to_float(chunk[84:88])]  # Zs in geographic coordinate system

    meta = {'software processing version': int(codecs.encode(chunk[88:89],
                                                             'hex'), 16),
            'software processing subversion': int(codecs.encode(chunk[89:90],
                                                                'hex'), 16),
            'data names': ['glat', 'glon', 'altitude', 'LT', 'mlat', 'mlon',
                           'MLT', 'ilat', 'L', 'glat_conj', 'glon_conj',
                           'glat_conj_N_110km', 'glon_conj_N_110km',
                           'glat_conj_S_110km', 'glon_conj_S_110km',
                           'mag_comp_1', 'mag_comp_2', 'mag_comp_3',
                           'proton_gyrofreq', 'Xs', 'Ys', 'Zs'],
            'data units': {'glat': 'degrees',
                           'glon': 'degrees',
                           'altitude': 'km',
                           'LT': 'h',
                           'mlat': 'degrees',
                           'mlon': 'degrees',
                           'MLT': 'h',
                           'ilat': 'degrees',
                           'L': 'Earth_Radii',
                           'glat_conj': 'degrees',
                           'glon_conj': 'degrees',
                           'glat_conj_N_110km': 'degrees',
                           'glon_conj_N_110km': 'degrees',
                           'glat_conj_S_110km': 'degrees',
                           'glon_conj_S_110km': 'degrees',
                           'mag_comp_1': 'nT',
                           'mag_comp_2': 'nT',
                           'mag_comp_3': 'nT',
                           'proton_gyrofreq': 'Hz',
                           'Xs': 'N/A',
                           'Ys': 'N/A',
                           'Zs': 'N/A'}}

    return data, meta


def load_attitude_parameters(fhandle):
    """ Load the attitude parameter block (block 1 for each time)

    Parameters
    ------------
    fhandle : (file handle)
        File handle

    Returns
    ----------
    data : list
        list of data values containing: matrix elements from satellite coord
        system to geographic coordinate system, matrix elements from geographic
        coordinate system to local geomagnetic coordinate system, quality
        index of attitude parameters.
    meta : dict
        Dictionary with meta data for keys: 'software processing version',
        'software processing subversion', 'data names', 'data units'

    """
    import codecs  # ensures encode is python 2/3 compliant

    block_size = 76  # Position block is 76 bytes
    chunk = fhandle.read(block_size)
    data = list()

    if len(chunk) != block_size:
        return data, dict()

    i = 0
    j = 1
    k = 1
    data_names = list()
    data_units = dict()
    while i < 72:
        data.append(bytes_to_float(chunk[i:i+4]))  # Matrix element

        # Save data name and units
        data_names.append("{:s}_{:d}{:d}".format("sat2geo" if i <= 32
                                                 else "geo2lgm", j, k))
        data_units[data_names[-1]] = 'unitless'

        # Cycle to next chunk of data
        i += 4

        # Determine the next matrix index
        if k == 3:
            k = 1
            j = j + 1 if j < 3 else 1
        else:
            k += 1

    # Save quality flag
    data.append(int(codecs.encode(chunk[72:74], 'hex'), 16))
    data_names.append('attitude_flag')
    data_units[data_names[-1]] = 'unitless'

    meta = {'software processing version': int(codecs.encode(chunk[74:75],
                                                             'hex'), 16),
            'software processing subversion': int(codecs.encode(chunk[75:76],
                                                                'hex'), 16),
            'data names': data_names, 'data units': data_units}

    return data, meta


def load_binary_file(fname, load_experiment_data):
    """ Load the binary data from a DEMETER file

    Parameters
    ------------
    fname : string
        Filename
    load_experiment_data : function
        Function to load experiment data, taking the file handle as input

    Returns
    ----------
    data : np.array
        Data from file stored in a numpy array
    meta : dict
        Meta data for file, including data names and units
    """

    data = list()
    meta = dict()

    with open(fname, "rb") as f:
        # Cycle through teach time, which consists of four blocks
        gdata, meta = load_general_header(f)

        while len(gdata) > 0:
            ldata, lmeta = load_location_parameters(f)
            adata, ameta = load_attitude_parameters(f)
            edata, emeta = load_experiment_data(f)

            # Combine and save the meta data
            if len(data) == 0:
                meta['data names'].extend(lmeta['data names'])
                meta['data units'].update(lmeta['data units'])
                meta['data names'].extend(ameta['data names'])
                meta['data units'].update(ameta['data units'])

                for ekey in emeta.keys():
                    if ekey == 'data names':
                        meta[ekey].extend(emeta[ekey])
                    elif ekey == 'data units':
                        meta[ekey].update(emeta[ekey])
                    else:
                        meta[ekey] = emeta[ekey]

            # Combine and save the data
            gdata.extend(ldata)
            gdata.extend(adata)
            gdata.extend(edata)
            data.append(gdata)

            # Cycle to the next time chunk
            gdata, _ = load_general_header(f)

        # Having read the file, close file handle and recast the data
        f.close()
        data = np.array(data)

    return data, meta


def set_metadata(name, meta_dict):
    """ Set metadata for each DEMETER instrument, using dict containing
    metadata

    Parameters
    ------------
    name : string
        DEMETER instrument name
    meta_dict : dict
        Dictionary containing metadata information and data attributes.  Data
        attributes are available in the keys 'data names' and 'data units'

    Returns
    ----------
    meta : pysat.Meta
        Meta class boject

    """

    # Define the acknowledgements and references
    ackn = {'iap': ' '.join(["This work is based on observations with the",
                             "plasma analyser IAP embarked on the satellite",
                             "DEMETER launched by CNES (Centre National",
                             "d'Etudes Spatiales). The author thanks J.J.",
                             "Berthelier the PI of this instrument for the use",
                             "of the data, and CDPP (Centre des Données de la",
                             "Physique des Plasmas) for the provision of these",
                             "data."]),
            'ice': ' '.join(["This work is based on observations with the",
                             "electric field instrument ICE embarked on the",
                             "satellite DEMETER launched by CNES (Centre",
                             "National d'Etudes Spatiales). The author thanks",
                             "J.J. Berthelier the PI of this instrument for",
                             "the use of the data, and CDPP (Centre des",
                             "Données de la Physique des Plasmas) for the",
                             "provision of these data."]),
            'imsc': ' '.join(["This work is based on observations with the",
                              "magnetic field instrument IMSC embarked on the",
                              "satellite DEMETER launched by CNES (Centre",
                              "National d'Etudes Spatiales). The author thanks",
                              "M. Parrot the PI of this instrument for the use",
                              "of the data, and CDPP (Centre des Données de la",
                              "Physique des Plasmas) for the provision of",
                              "these data."]),
            'rnf': ' '.join(["This work is based on observations with the",
                             "neural network RNF embarked on the satellite",
                             "DEMETER launched by CNES (Centre National",
                             "d'Etudes Spatiales). The author thanks J.L.",
                             "Pinçon the PI of this instrument for the use",
                             "of the data, and CDPP (Centre des Données de",
                             "la Physique des Plasmas) for the provision of",
                             "these data."]),
            'idp': ' '.join(["This work is based on observations with the",
                             "particle spectrometer instrument IDP embarked on",
                             "the satellite DEMETER launched by CNES (Centre",
                             "National d'Etudes Spatiales). The author thanks",
                             "J.A. Sauvaud the PI of this instrument for the",
                             "use of the data, and CDPP (Centre des Données",
                             "de la Physique des Plasmas) for the provision of",
                             "these data."]),
            'isl': ' '.join(["This work is based on observations with the",
                             "Langmuir probe ISL embarked on the satellite",
                             "DEMETER launched by CNES (Centre National",
                             "d'Etudes Spatiales). The author thanks J.P.",
                             "Lebreton the PI of this instrument for the use",
                             "of the data, and CDPP (Centre des Données de la",
                             "Physique des Plasmas) for the provision of these",
                             "data."])}

    refs = {'iap': ' '.join(['Berthelier at al., 2006. IAP, the thermal plasma',
                             'analyzer on DEMETER, Planet. and Space Sci.,',
                             '54(5), pp 487-501.'])}

    if name not in refs.keys():
        refs[name] = 'Instrument reference information available at ' + \
            'https://demeter.cnes.fr/en/DEMETER/A_publications.htm'

    # Define the long-form names for non-instrument specific data
    long_name = {'P_field': 'P field',
                 'epoch_time': 'Number of days from 01/01/1950',
                 'time_of_day': 'Number of milliseconds in the day',
                 'UT': 'Universal Time of the first point of the data array',
                 'orbit_number': 'Orbit number',
                 'orbit_type': 'Sub-orbit type: False=downward, True=upward',
                 'glat': 'Geocentric Latitude',
                 'glon': 'Geocentric Longitude',
                 'altitude': 'Altitude',
                 'LT': 'Local Time',
                 'mlat': 'Geomagnetic Latitude',
                 'mlon': 'Geomagnetic Longitude',
                 'MLT': 'Magnetic Local Time',
                 'ilat': 'Invarient Latitude',
                 'L': 'Mc Ilwain Parameter L',
                 'glat_conj': ' '.join(['Geocentric latitude of the conjugate',
                                        'point at the satellite altitude']),
                 'glon_conj': ' '.join(['Geocentric longitude of the conjugate',
                                        'point at the satellite altitude']),
                 'glat_conj_N_110km': ' '.join(['Geocentric latitude of North',
                                                'conjugate point at altitude',
                                                '110 km']),
                 'glon_conj_N_110km': ' '.join(['Geocentric longitude of North',
                                                'conjugate point at altitude',
                                                '110 km']),
                 'glat_conj_S_110km': ' '.join(['Geocentric latitude of South',
                                                'conjugate point at altitude',
                                                '110 km']),
                 'glon_conj_S_110km': ' '.join(['Geocentric longitude of South',
                                                'conjugate point at altitude',
                                                '110 km']),
                 'mag_comp_1': ' '.join(['Component of the magnetic field',
                                         'model at the satellite point']),
                 'mag_comp_2': ' '.join(['Component of the magnetic field',
                                         'model at the satellite point']),
                 'mag_comp_3': ' '.join(['Component of the magnetic field',
                                         'model at the satellite point']),
                 'proton_gyrofreq': 'Proton gyrofrequency at satellite point',
                 'Xs': 'Solar position in geographic coordinate system',
                 'Ys': 'Solar position in geographic coordinate system',
                 'Zs': 'Solar position in geographic coordinate system',
                 'sat2geo_11': ' '.join(['Conversion matrix from satellite to',
                                         'geographic coordinate system']),
                 'sat2geo_12': ' '.join(['Conversion matrix from satellite to',
                                         'geographic coordinate system']),
                 'sat2geo_13': ' '.join(['Conversion matrix from satellite to',
                                         'geographic coordinate system']),
                 'sat2geo_21': ' '.join(['Conversion matrix from satellite to',
                                         'geographic coordinate system']),
                 'sat2geo_22': ' '.join(['Conversion matrix from satellite to',
                                         'geographic coordinate system']),
                 'sat2geo_23': ' '.join(['Conversion matrix from satellite to',
                                         'geographic coordinate system']),
                 'sat2geo_31': ' '.join(['Conversion matrix from satellite to',
                                         'geographic coordinate system']),
                 'sat2geo_32': ' '.join(['Conversion matrix from satellite to',
                                         'geographic coordinate system']),
                 'sat2geo_33': ' '.join(['Conversion matrix from satellite to',
                                         'geographic coordinate system']),
                 'geo2lgm_11': ' '.join(['Conversion matrix from geographic to',
                                         'geomagnetic coordinate system']),
                 'geo2lgm_12': ' '.join(['Conversion matrix from geographic to',
                                         'geomagnetic coordinate system']),
                 'geo2lgm_13': ' '.join(['Conversion matrix from geographic to',
                                         'geomagnetic coordinate system']),
                 'geo2lgm_21': ' '.join(['Conversion matrix from geographic to',
                                         'geomagnetic coordinate system']),
                 'geo2lgm_22': ' '.join(['Conversion matrix from geographic to',
                                         'geomagnetic coordinate system']),
                 'geo2lgm_23': ' '.join(['Conversion matrix from geographic to',
                                         'geomagnetic coordinate system']),
                 'geo2lgm_31': ' '.join(['Conversion matrix from geographic to',
                                         'geomagnetic coordinate system']),
                 'geo2lgm_32': ' '.join(['Conversion matrix from geographic to',
                                         'geomagnetic coordinate system']),
                 'geo2lgm_33': ' '.join(['Conversion matrix from geographic to',
                                         'geomagnetic coordinate system']),
                 'attitude_flag': 'Quality index of attitude parameters',
                 'status_flag_00': 'Housekeeping and status',
                 'status_flag_01': 'Housekeeping and status',
                 'status_flag_02': 'Housekeeping and status',
                 'status_flag_03': 'Housekeeping and status',
                 'status_flag_04': 'Housekeeping and status',
                 'status_flag_05': 'Housekeeping and status',
                 'status_flag_06': 'Housekeeping and status',
                 'status_flag_07': 'Housekeeping and status',
                 'status_flag_08': 'Housekeeping and status',
                 'status_flag_09': 'Housekeeping and status',
                 'status_flag_10': 'Housekeeping and status',
                 'status_flag_11': 'Housekeeping and status',
                 'status_flag_12': 'Housekeeping and status',
                 'status_flag_13': 'Housekeeping and status',
                 'status_flag_14': 'Housekeeping and status',
                 'status_flag_15': 'Housekeeping and status',
                 'status_flag_16': 'Housekeeping and status',
                 'status_flag_17': 'Housekeeping and status',
                 'status_flag_18': 'Housekeeping and status',
                 'status_flag_19': 'Housekeeping and status',
                 'status_flag_20': 'Housekeeping and status',
                 'status_flag_21': 'Housekeeping and status',
                 'status_flag_22': 'Housekeeping and status',
                 'status_flag_23': 'Housekeeping and status',
                 'status_flag_24': 'Housekeeping and status',
                 'status_flag_25': 'Housekeeping and status',
                 'status_flag_26': 'Housekeeping and status',
                 'status_flag_27': 'Housekeeping and status',
                 'status_flag_28': 'Housekeeping and status',
                 'status_flag_29': 'Housekeeping and status',
                 'status_flag_30': 'Housekeeping and status',
                 'status_flag_31': 'Housekeeping and status', }
    long_inst = {'iap': {'time_resolution': 'Time resolution',
                         'H+_density': 'H+ density',
                         'He+_density': 'He+ density',
                         'O+_density': 'O+ density',
                         'Ion_temperature': 'Ion temperature',
                         'iv_Oz': 'Ion velocity along the satellite z axis',
                         'iv_negOz_angle': ' '.join(['Angle between the ion',
                                                     'velocity and -z axis',
                                                     '(ram direction) of',
                                                     'satellite']),
                         'iv_xOy_Ox_angle': ' '.join(['Angle between',
                                                      'projection of the ion',
                                                      'velocity on the x-y',
                                                      'plane and satellite',
                                                      'x axis']),
                         'satellite_potential': 'Satellite potential'}}

    if name not in long_inst.keys():
        logger.warning('no long-form names available for {:s}'.format(name))

        long_inst[name] = {nn: nn for nn in meta_dict['data names']}

    # Initialise the meta data
    meta = pysat.Meta()
    for cc in meta_dict['data names']:
        # Determine the long instrument name
        if cc in long_inst[name].keys():
            ll = long_inst[name][cc]
        else:
            ll = long_name[cc]

        # Assign the data units, long names, acknowledgements, and references
        meta[cc] = {'units': meta_dict['data units'][cc], 'long_name': ll}

    # Set the remaining metadata
    meta_dict['acknowledgements'] = ackn[name]
    meta_dict['reference'] = refs[name]
    mkeys = list(meta_dict.keys())
    mkeys.pop(mkeys.index('data names'))
    mkeys.pop(mkeys.index('data units'))

    meta.info = {cc: meta_dict[cc] for cc in mkeys}

    return meta



import pysat
"""
Creates a constellation from NASA ICON instrumentation
"""


lang = pysat.Instrument(platform='de2', name='lang')
nacs = pysat.Instrument(platform='de2', name='nacs')
rpa = pysat.Instrument(platform='de2', name='rpa')
wats = pysat.Instrument(platform='de2', name='wats')

instruments = [lang, nacs, rpa, wats]



"""
tests the pysat utils.stats area
"""
import numpy as np
from scipy import stats as scistats
import warnings
from pysat.utils import stats as pystats


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""

        # Add testing data for circular statistics
        self.test_angles = np.array([340.0, 348.0, 358.9, 0.5, 5.0, 9.87])
        self.test_nan = [340.0, 348.0, 358.9, 0.5, 5.0, 9.87, np.nan]
        self.circ_kwargs = {"high": 360.0, "low": 0.0}

    def teardown(self):
        """Runs after every method to clean up previous testing."""

        del self.test_angles, self.test_nan, self.circ_kwargs

    def test_circmean(self):
        """ Test custom circular mean."""

        ref_mean = scistats.circmean(self.test_angles, **self.circ_kwargs)
        test_mean = pystats.nan_circmean(self.test_angles, **self.circ_kwargs)

        assert ref_mean == test_mean

    def test_circmean_nan(self):
        """ Test custom circular mean with NaN."""

        ref_mean = scistats.circmean(self.test_angles, **self.circ_kwargs)
        ref_nan = scistats.circmean(self.test_nan, **self.circ_kwargs)
        test_nan = pystats.nan_circmean(self.test_nan, **self.circ_kwargs)

        assert np.isnan(ref_nan)
        assert ref_mean == test_nan

    def test_circstd(self):
        """ Test custom circular std."""

        ref_std = scistats.circstd(self.test_angles, **self.circ_kwargs)
        test_std = pystats.nan_circstd(self.test_angles, **self.circ_kwargs)

        assert ref_std == test_std

    def test_circstd_nan(self):
        """ Test custom circular std with NaN."""

        ref_std = scistats.circstd(self.test_angles, **self.circ_kwargs)
        ref_nan = scistats.circstd(self.test_nan, **self.circ_kwargs)
        test_nan = pystats.nan_circstd(self.test_nan, **self.circ_kwargs)

        assert np.isnan(ref_nan)
        assert ref_std == test_nan


class TestDeprecation():

    def setup(self):
        """Runs before every method to create a clean testing setup"""
        warnings.simplefilter("always")

    def teardown(self):
        """Runs after every method to clean up previous testing"""

    def test_deprecation_warning_median1d(self):
        """Test if median1D in stats is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                pystats.median1D(None, [0, 24, 2], 'slt', 'slt')
            except AttributeError:
                # Setting inst to None should produce a AttributeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_deprecation_warning_circmean(self):
        """Test if circmean in stats is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                pystats.nan_circmean(None)
            except TypeError:
                # Setting input to None should produce a TypeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_deprecation_warning_circstd(self):
        """Test if circstd in stats is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                pystats.nan_circstd(None)
            except TypeError:
                # Setting input to None should produce a TypeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning



"""
pysat.constellations is a pysat module that provides
the interface for pysat to manage and analyze data
from multiple pysat instrument.  Each instrument
is contained within a subpackage of the pysat.instruments
package.
"""

__all__ = ['de2', 'icon']
